import os, struct, math

# Item names as found in BATTLE.BIN
items = {
    0x01: "Dagger",
    0x02: "Mythril Knife",
    0x03: "Blind Knife",
    0x04: "Mage Masher",
    0x05: "Platina Dagger",
    0x06: "Main Gauche",
    0x07: "Orichalcum",
    0x08: "Assassin Dagger",
    0x09: "Air Knife",
    0x0A: "Zorlin Shape",
    0x0B: "Hidden Knife",
    0x0C: "Ninja Knife",
    0x0D: "Short Edge",
    0x0E: "Ninja Edge",
    0x0F: "Spell Edge",
    0x10: "Sasuke Knife",
    0x11: "Iga Knife",
    0x12: "Koga Knife",
    0x13: "Broad Sword",
    0x14: "Long Sword",
    0x15: "Iron Sword",
    0x16: "Mythril Sword",
    0x17: "Blood Sword",
    0x18: "Coral Sword",
    0x19: "Ancient Sword",
    0x1A: "Sleep Sword",
    0x1B: "Platinum Sword",
    0x1C: "Diamond Sword",
    0x1D: "Ice Brand",
    0x1E: "Rune Blade",
    0x1F: "Nagrarock",
    0x20: "Materia Blade",
    0x21: "Defender",
    0x22: "Save the Queen",
    0x23: "Excalibur",
    0x24: "Ragnarok",
    0x25: "Chaos Blade",
    0x26: "Asura Knife",
    0x27: "Koutetsu Knife",
    0x28: "Bizen Boat",
    0x29: "Murasame",
    0x2A: "Heaven's Cloud",
    0x2B: "Kiyomori",
    0x2C: "Muramasa",
    0x2D: "Kikuichimoji",
    0x2E: "Masamune",
    0x2F: "Chirijiraden",
    0x30: "Battle Axe",
    0x31: "Giant Axe",
    0x32: "Slasher",
    0x33: "Rod",
    0x34: "Thunder Rod",
    0x35: "Flame Rod",
    0x36: "Ice Rod",
    0x37: "Poison Rod",
    0x38: "Wizard Rod",
    0x39: "Dragon Rod",
    0x3A: "Faith Rod",
    0x3B: "Oak Staff",
    0x3C: "White Staff",
    0x3D: "Healing Staff",
    0x3E: "Rainbow Staff",
    0x3F: "Wizard Staff",
    0x40: "Gold Staff",
    0x41: "Mace of Zeus",
    0x42: "Sage Staff",
    0x43: "Flail",
    0x44: "Flame Whip",
    0x45: "Morning Star",
    0x46: "Scorpion Tail",
    0x47: "Romanda Gun",
    0x48: "Mythril Gun",
    0x49: "Stone Gun",
    0x4A: "Blaze Gun",
    0x4B: "Glacier Gun",
    0x4C: "Blast Gun",
    0x4D: "Bow Gun",
    0x4E: "Night Killer",
    0x4F: "Cross Bow",
    0x50: "Poison Bow",
    0x51: "Hunting Bow",
    0x52: "Gastrafitis",
    0x53: "Long Bow",
    0x54: "Silver Bow",
    0x55: "Ice Bow",
    0x56: "Lightning Bow",
    0x57: "Windslash Bow",
    0x58: "Mythril Bow",
    0x59: "Ultimus Bow",
    0x5A: "Yoichi Bow",
    0x5B: "Perseus Bow",
    0x5C: "Ramia Harp",
    0x5D: "Bloody Strings",
    0x5E: "Fairy Harp",
    0x5F: "Battle Dict",
    0x60: "Monster Dict",
    0x61: "Papyrus Plate",
    0x62: "Madlemgen",
    0x63: "Javelin",
    0x64: "Spear",
    0x65: "Mythril Spear",
    0x66: "Partisan",
    0x67: "Oberisk",
    0x68: "Holy Lance",
    0x69: "Dragon Whisker",
    0x6A: "Javelin",
    0x6B: "Cypress Rod",
    0x6C: "Battle Bamboo",
    0x6D: "Musk Rod",
    0x6E: "Iron Fan",
    0x6F: "Gokuu Rod",
    0x70: "Ivory Rod",
    0x71: "Octagon Rod",
    0x72: "Whale Whisker",
    0x73: "C Bag",
    0x74: "FS Bag",
    0x75: "P Bag",
    0x76: "H Bag",
    0x77: "Persia",
    0x78: "Cashmere",
    0x79: "Ryozan Silk",
    0x7A: "Shuriken",
    0x7B: "Magic Shuriken",
    0x7C: "Yagyu Darkness",
    0x7D: "Fire Ball",
    0x7E: "Water Ball",
    0x7F: "Lightning Ball",
    0x80: "Escutcheon",
    0x81: "Buckler",
    0x82: "Bronze Shield",
    0x83: "Round Shield",
    0x84: "Mythril Shield",
    0x85: "Gold Shield",
    0x86: "Ice Shield",
    0x87: "Flame Shield",
    0x88: "Aegis Shield",
    0x89: "Diamond Shield",
    0x8A: "Platina Shield",
    0x8B: "Crystal Shield",
    0x8C: "Genji Shield",
    0x8D: "Kaiser Plate",
    0x8E: "Venetian Shield",
    0x8F: "Escutcheon",
    0x90: "Leather Helmet",
    0x91: "Bronze Helmet",
    0x92: "Iron Helmet",
    0x93: "Barbuta",
    0x94: "Mythril Helmet",
    0x95: "Gold Helmet",
    0x96: "Cross Helmet",
    0x97: "Diamond Helmet",
    0x98: "Platina Helmet",
    0x99: "Circlet",
    0x9A: "Crystal Helmet",
    0x9B: "Genji Helmet",
    0x9C: "Grand Helmet",
    0x9D: "Leather Hat",
    0x9E: "Feather Hat",
    0x9F: "Red Hood",
    0xA0: "Headgear",
    0xA1: "Triangle Hat",
    0xA2: "Green Beret",
    0xA3: "Twist Headband",
    0xA4: "Holy Miter",
    0xA5: "Black Hood",
    0xA6: "Golden Hairpin",
    0xA7: "Flash Hat",
    0xA8: "Thief Hat",
    0xA9: "Cachusha",
    0xAA: "Barette",
    0xAB: "Ribbon",
    0xAC: "Leather Armor",
    0xAD: "Linen Cuirass",
    0xAE: "Bronze Armor",
    0xAF: "Chain Mail",
    0xB0: "Mythril Armor",
    0xB1: "Plate Mail",
    0xB2: "Gold Armor",
    0xB3: "Diamond Armor",
    0xB4: "Platina Armor",
    0xB5: "Carabini Mail",
    0xB6: "Crystal Mail",
    0xB7: "Genji Armor",
    0xB8: "Reflect Mail",
    0xB9: "Maximillian",
    0xBA: "Clothes",
    0xBB: "Leather Outfit",
    0xBC: "Leather Vest",
    0xBD: "Chain Vest",
    0xBE: "Mythril Vest",
    0xBF: "Adaman Vest",
    0xC0: "Wizard Outfit",
    0xC1: "Brigandine",
    0xC2: "Judo Outfit",
    0xC3: "Power Sleeve",
    0xC4: "Earth Clothes",
    0xC5: "Secret Clothes",
    0xC6: "Black Costume",
    0xC7: "Rubber Costume",
    0xC8: "Linen Robe",
    0xC9: "Silk Robe",
    0xCA: "Wizard Robe",
    0xCB: "Chameleon Robe",
    0xCC: "White Robe",
    0xCD: "Black Robe",
    0xCE: "Light Robe",
    0xCF: "Robe of Lords",
    0xD0: "Battle Boots",
    0xD1: "Spike Shoes",
    0xD2: "Germinas Boots",
    0xD3: "Rubber Shoes",
    0xD4: "Feather Boots",
    0xD5: "Sprint Shoes",
    0xD6: "Red Shoes",
    0xD7: "Power Wrist",
    0xD8: "Genji Gauntlet",
    0xD9: "Magic Gauntlet",
    0xDA: "Bracer",
    0xDB: "Reflect Ring",
    0xDC: "Defense Ring",
    0xDD: "Magic Ring",
    0xDE: "Cursed Ring",
    0xDF: "Angel Ring",
    0xE0: "Diamond Armlet",
    0xE1: "Jade Armlet",
    0xE2: "108 Gems",
    0xE3: "NKai Armlet",
    0xE4: "Defense Armlet",
    0xE5: "Small Mantle",
    0xE6: "Leather Mantle",
    0xE7: "Wizard Mantle",
    0xE8: "Elf Mantle",
    0xE9: "Dracula Mantle",
    0xEA: "Feather Mantle",
    0xEB: "Vanish Mantle",
    0xEC: "Chantage",
    0xED: "Cherche",
    0xEE: "Setiemson",
    0xEF: "Salty Rage",
    0xF0: "Potion",
    0xF1: "HiPotion",
    0xF2: "XPotion",
    0xF3: "Ether",
    0xF4: "HiEther",
    0xF5: "Elixir",
    0xF6: "Antidote",
    0xF7: "Eye Drop",
    0xF8: "Echo Grass",
    0xF9: "Maiden's Kiss",
    0xFA: "Soft",
    0xFB: "Holy Water",
    0xFC: "Remedy",
    0xFD: "Phoenix Down"
}

class BattleBin(file):

    def __init__(self, file_name):
        file.__init__(self, file_name, 'rb')
        self.toc_traps = 0x8ee74

    def read_trap(self, map_num):
        self.seek(self.toc_traps + map_num * 0x10)
        traps = []
        for trapnum in range(4):
            trap = {}
            trapdata = struct.unpack('<BBBB', self.read(4))
            trap['x'] = trapdata[0] >> 4
            trap['z'] = trapdata[0] & 0xf
            trap['type'] = trapdata[1]
            try:
                trap['rare'] = items[trapdata[2]]
                trap['common'] = items[trapdata[3]]
                print '(%02u, %02u): %s / %s' % (trap['x'], trap['z'], trap['common'], trap['rare'])
                traps.append(trap)
            except KeyError:
                pass
        return traps


class AttackOut(file):

    def __init__(self, file_name):
        file.__init__(self, file_name, 'rb')
        self.toc_placements = 0xbbd4
        self.toc_scenarios = 0x10938

    def read_placements(self, map_num):
        def rotate_bmp(bmp, times):
            for t in range(times):
                bmp.reverse()
                new_bmp = []
                for i in range(5):
                    new_row = []
                    for row in bmp:
                        new_row.append(row[i])
                    new_bmp.append(new_row)
                bmp = new_bmp
            return bmp
        def flip_x(bmp):
            for row in bmp:
                row.reverse()
            return bmp
        def flip_z(bmp):
            bmp.reverse()
            return bmp
        placements = []
        self.seek(self.toc_placements)
        for i in range(768):
            b = [0, 0, 0, 0]
            bits = []
            bitmap = []
            (b[0], b[1], b[2], b[3], px, pz, orient, chars, map_id, place_id) = \
                    struct.unpack('<BBBBbbxBBBH', self.read(12))
            if not map_id == map_num:
                continue
            for byte in b:
                bits.append((byte >> 0) & 0x1)
                bits.append((byte >> 1) & 0x1)
                bits.append((byte >> 2) & 0x1)
                bits.append((byte >> 3) & 0x1)
                bits.append((byte >> 4) & 0x1)
                bits.append((byte >> 5) & 0x1)
                bits.append((byte >> 6) & 0x1)
                bits.append((byte >> 7) & 0x1)
            j = 0
            rotate = (orient >> 4) & 0xf
            face = (orient >> 0) & 0xf
            for z in range(-2, 2 + 1):
                row = []
                for x in range(-2, 2 + 1):
                    row.append(bits[j])
                    j += 1
                bitmap.append(row)
            bitmap = flip_x(bitmap)
            bitmap = rotate_bmp(bitmap, 1)
            bitmap = rotate_bmp(bitmap, rotate)
            if i < 256: when = 'random'
            elif i < 512: when = 'story'
            else: when = 'other'
            placements.append({'bitmap': bitmap, 'x': px, 'z': pz,
                'face': face, 'rotate': rotate, 'when': when, 'id': place_id})
        return placements

    def read_scenarios(self):
        self.seek(self.toc_scenarios)
        scens = []
        for i in range(512):
            scen = {}
            #                      012345678         9
            line = struct.unpack('<HBBBBBHHHxxxxxxxxxH', self.read(24))
            scen['event'] = line[9]
            if i > 0 and scen['event'] == 0: continue
            scen['map_number'] = line[1]
            scen['arrangement'] = line[2] >> 4
            scen['weather'] = line[2] & 0xF
            scen['day_night'] = line[3]
            scen['entd'] = line[6]
            scen['squad1'] = line[7]
            scen['squad2'] = line[8]
            scens.append(scen)
        return scens


class Entd(file):
    def __init__(self, map_path, id):
        offset = id * 40 * 16
        file_num = math.floor(offset / 81920.0) + 1
        offset = offset % 81920
        path = os.path.join(map_path, '..', 'BATTLE', 'ENTD%u.ENT' % file_num)
        path = os.path.normpath(path)
        print 'Reading unit positions from %s offset 0x%x' % (path, offset)
        file.__init__(self, path, 'rb')
        self.seek(offset)

    def read_entds(self):
        entds = []
        for i in range(16):
            entd = {}
            self.seek(24, 1)
            line = struct.unpack('BBBB', self.read(4))
            if line[0] & 0x80:
                entd['always'] = True
                entd['enemy'] = (line[0] & 0x10) == 0x10
                control = (line[0] & 0x8) == 0x8
                immortal = (line[0] & 0x4) == 0x4
                entd['guest'] = not entd['enemy'] and not control and immortal
                entd['x'] = line[1]
                entd['y'] = (line[3] & 0x80) >> 7
                entd['z'] = line[2]
                entd['face'] = line[3] & 0x7f
                entds.append(entd)
            self.seek(12, 1)
        return entds
