#!/usr/bin/python
# coding: utf-8

# Final Fantasy Tactics Map2GL Map Viewer
# Don Laursen, 2006-2008
#
# Thanks to the following for their work on OpenGL tutorials, from which this
# program grew:
# Richard Campbell
# John Ferguson
# Tarn Weisner Burton

import os, sys
sys.path.insert(0, os.path.join(sys.prefix, "pyopengl-3.0.0b6-py2.5.egg"))
sys.path.insert(0, os.path.join(sys.prefix, "setuptools-0.6c9-py2.5.egg"))
sys.path.insert(0, os.path.join(sys.prefix, "PIL-1.1.6-py2.5-win32.egg"))
sys.path.insert(0, os.path.join(sys.prefix, "numpy-1.2.1-py2.5-win32.egg"))
import wx, wx.glcanvas
import fft.map.viewer, fft.map.texedit

# Global variables
(default_w, default_h) = (800, 600)

ID_INFO = 100
ID_CAMERA = 101
ID_SCENARIO = 102
ID_TEXEDIT = 103
ID_INDEX_CHOOSER = 1000
ID_AZIMUTH_TEXT = 1100
ID_AZIMUTH_SLIDE = ID_AZIMUTH_TEXT + 1
ID_AZIMUTH_AUTO = ID_AZIMUTH_TEXT + 2
ID_AZIMUTH_0 = ID_AZIMUTH_TEXT + 3
ID_AZIMUTH_M45 = ID_AZIMUTH_TEXT + 4
ID_AZIMUTH_P45 = ID_AZIMUTH_TEXT + 5
ID_ALTITUDE_TEXT = 1150
ID_ALTITUDE_SLIDE = ID_ALTITUDE_TEXT + 1
ID_ALTITUDE_0 = ID_ALTITUDE_TEXT + 2
ID_ALTITUDE_26 = ID_ALTITUDE_TEXT + 3
ID_ALTITUDE_30 = ID_ALTITUDE_TEXT + 4
ID_ALTITUDE_38 = ID_ALTITUDE_TEXT + 5
ID_ALTITUDE_45 = ID_ALTITUDE_TEXT + 6
ID_ALTITUDE_90 = ID_ALTITUDE_TEXT + 7
ID_MAGNIFY_TEXT = 1200
ID_MAGNIFY_SLIDE = ID_MAGNIFY_TEXT + 1
ID_MAGNIFY_AUTO = ID_MAGNIFY_TEXT + 2
ID_MAGNIFY_100 = ID_MAGNIFY_TEXT + 3
ID_FOV_TEXT = 1250
ID_FOV_SLIDE = ID_FOV_TEXT + 1
ID_FOV_ISOMETRIC = ID_FOV_TEXT + 2
ID_FOV_27 = ID_FOV_TEXT + 3
ID_SCENARIO_CHOOSER = 2000
ID_EXPORT = 3001
ID_IMPORT = 3002
ID_SHOW_USED_AREAS = 3051
ID_PALETTE_RADIO_0 = 3100
ID_COLOR_BUTTON_0 = 3120
ID_COLOR_LABEL_0 = 3410
ID_COLOR_SLIDER_0 = 3420
ID_COLOR_INPUT_0 = 3430

class ViewerCanvas(wx.glcanvas.GLCanvas):
    def __init__(self, parent, name, size):
        attribs = [wx.glcanvas.WX_GL_RGBA, wx.glcanvas.WX_GL_DOUBLEBUFFER, wx.glcanvas.WX_GL_DEPTH_SIZE, 24]
        wx.glcanvas.GLCanvas.__init__(self, parent, -1, name = name, size = size, attribList=attribs)
        self.timer = wx.Timer(self)
        wx.EVT_PAINT(self, self.on_paint)
        wx.EVT_SIZE(self, self.on_resize)
        self.Bind(wx.EVT_TIMER, self.on_timer)
        self.timer.Start(5)
        self.mouse_drag = False
        self.init = False
        wx.EVT_MOTION(self, self.on_motion)
        wx.EVT_LEFT_DOWN(self, self.on_left_down)
        wx.EVT_LEFT_UP(self, self.on_left_up)
        wx.EVT_ERASE_BACKGROUND(self, self.on_erase_bg)
        self.click_begin_azimuth = None
        self.click_begin_altitude = None
        self.click_begin_x = None
        self.click_begin_y = None

    def on_timer(self, event):
        if self.init:
            self.on_paint(None)

    def on_resize(self, event):
        size = self.GetClientSize()
        (self.W, self.H) = (size.width, size.height)
        if self.GetContext():
            self.SetCurrent()
            app.viewer.gl_set_viewport(size.width, size.height)
        event.Skip()

    def on_motion(self, event):
        global angle, altitude
        if self.mouse_drag:
            (x, y) = event.GetPosition()
            app.viewer.angle = self.click_begin_azimuth + (x - self.click_begin_x) * 360.0 / self.W
            app.viewer.altitude = self.click_begin_altitude + (y - self.click_begin_y) * 180.0 / self.H
            app.viewer.mod_angles()

    def on_left_down(self, event):
        (x, y) = event.GetPosition()
        self.mouse_drag = True
        self.click_begin_x = x
        self.click_begin_y = y
        self.click_begin_azimuth = app.viewer.angle
        self.click_begin_altitude = app.viewer.altitude
        app.viewer.angleinc = 0

    def on_left_up(self, event):
        app.viewer.mod_angles()
        self.mouse_drag = False

    def on_paint(self, event):
        if event:
            dc = wx.PaintDC(self)
        self.SetCurrent()
        if not self.init:
            self.init = True
            app.viewer.gl_initialize()
            size = self.GetClientSize()
            (self.W, self.H) = (size.width, size.height)
            app.viewer.gl_set_viewport(self.W, self.H)
        self.draw_gl_stuff()

    def draw_gl_stuff(self):
        if app.viewer.map_is_loaded:
            app.viewer.gl_draw()
            self.SwapBuffers()
            app.camera_window.on_values_changed()

    def on_erase_bg(self, event):
        # Prevent flickering in Windows
        pass


class ViewerWindow(wx.Frame):
    def __init__(self, parent, ID, title):
        wx.Frame.__init__(self, parent, ID, title,
                wx.DefaultPosition)
        self.CreateStatusBar()
        self.SetStatusText("Final Fantasy Tactics map viewer")

        menuBar = wx.MenuBar()
        menu = wx.Menu()
        menu.Append(wx.ID_OPEN, "&Open", "Choose a map file to view")
        menu.AppendCheckItem(ID_SCENARIO, "&Scenario Chooser", "Choose a scenario to view")
        menu.AppendCheckItem(ID_TEXEDIT, "&Texture Editor", "Export texture as PNG and then re-import it")
        menu.AppendSeparator()
        menu.Append(wx.ID_EXIT, "&Quit", "Quit map2gl")
        menuBar.Append(menu, "&File")

        menu = wx.Menu()
        menu.AppendCheckItem(ID_INFO, "&Info", "Show information window")
        menu.AppendCheckItem(ID_CAMERA, "&Camera", "Show camera window")
        menuBar.Append(menu, "&Window")

        self.SetMenuBar(menuBar)

        self.SetClientSize((default_w, default_h))

        wx.EVT_MENU(self, wx.ID_OPEN, self.open_file_dialog)
        wx.EVT_MENU(self, ID_SCENARIO, self.toggle_scenario_window)
        wx.EVT_MENU(self, ID_TEXEDIT, self.toggle_texedit_window)
        wx.EVT_MENU(self, wx.ID_EXIT,  self.app_exit)
        wx.EVT_MENU(self, ID_INFO,  self.toggle_info_window)
        wx.EVT_MENU(self, ID_CAMERA,  self.toggle_camera_window)
        wx.EVT_MOVE(self, self.on_move)

    def open_file_dialog(self, event):
        dlg = wx.FileDialog(self, "Choose a GNS file", wildcard = 'GNS Files (*.GNS)|*.GNS;*.gns')
        if not app.map_path == None:
            dlg.SetDirectory(app.map_path)
        if dlg.ShowModal() == wx.ID_OK:
            file_spec = dlg.GetPath()
            self.open_file(file_spec)
        dlg.Destroy()

    def open_file(self, file_spec):
        print 'Loading:', file_spec
        app.map_path = os.path.dirname(file_spec)
        app.map_number = int(file_spec[-7:-4])
        app.viewer.load_traps(app.map_path, app.map_number)
        app.viewer.load_placements(app.map_path, app.map_number)
        app.viewer.entds = []
        app.viewer.load_gns(file_spec)
        if app.texedit_window.IsShown():
            app.texedit_window.canvas.SetCurrent()
            app.texedit.load_map(app.viewer.map, app.texedit_window.canvas.init)
            app.texedit_window.set_button_colors()
            if app.texedit_window.canvas.init:
                app.texedit_window.canvas.on_paint(None)
        app.gns_index = 0
        app.info_window.on_map_loaded()
        if app.scenarios == None:
            app.get_scenarios()

    def toggle_scenario_window(self, event):
        if event.Checked():
            app.scenario_window.Show(True)
        else:
            app.scenario_window.Show(False)

    def toggle_texedit_window(self, event):
        if event.Checked():
            app.texedit_window.Show(True)
            app.texedit_window.canvas.SetCurrent()
            app.texedit_window.canvas.on_paint(None)
            app.texedit.load_map(app.viewer.map, app.texedit_window.canvas.init)
            app.texedit_window.set_button_colors()
            app.texedit_window.canvas.on_paint(None)
        else:
            app.texedit_window.Show(False)

    def toggle_info_window(self, event):
        if event.Checked():
            app.info_window.position_self()
            app.info_window.Show(True)
        else:
            app.info_window.Show(False)

    def toggle_camera_window(self, event):
        if event.Checked():
            app.camera_window.Show(True)
        else:
            app.camera_window.Show(False)

    def on_move(self, event):
        try:
            app.info_window.position_self()
        except AttributeError:
            pass

    def app_exit(self, event):
        self.Close(True)


class InfoWindow(wx.Frame):
    def __init__(self, parent, ID, title):
        notstyle = wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX | wx.RESIZE_BORDER | wx.CLOSE_BOX
        wx.Frame.__init__(self, parent, ID, title,
                wx.DefaultPosition, wx.Size(250, default_h),
                wx.DEFAULT_FRAME_STYLE ^ notstyle)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.index_chooser = wx.ListBox(self, ID_INDEX_CHOOSER)
        sizer.Add(self.index_chooser, 1, wx.EXPAND)
        wx.EVT_LISTBOX_DCLICK(self, ID_INDEX_CHOOSER, self.on_list_dclick)
        self.SetSizer(sizer)

    def on_map_loaded(self):
        app.viewer_window.SetStatusText("%03u %s" % (app.map_number, fft.map.names[app.map_number]))
        index_choices = [app.viewer.map.stringify_index(i) for i in app.viewer.map.all_indexes]
        self.index_chooser.Set(index_choices)
        self.index_chooser.SetSelection(app.gns_index)
        self.Update()

    def on_list_dclick(self, event):
        app.gns_index = event.GetInt()
        gns_filespec = os.path.join(app.map_path, 'MAP%03i.GNS' % (app.map_number))
        app.viewer.load_gns(gns_filespec, app.gns_index)
        if app.texedit_window.IsShown():
            app.texedit_window.canvas.SetCurrent()
            app.texedit.load_map(app.viewer.map, app.texedit_window.canvas.init)
            app.texedit_window.set_button_colors()
            if app.texedit_window.canvas.init:
                app.texedit_window.canvas.on_paint(None)
        self.index_chooser.SetSelection(app.gns_index)

    def position_self(self):
        ppos = self.Parent.GetPosition()
        psize = self.Parent.GetSize()
        self.Move((ppos[0] + psize[0] + 10, ppos[1]))


class ScenarioWindow(wx.Frame):
    def __init__(self, parent, ID, title):
        notstyle = wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX | wx.RESIZE_BORDER | wx.CLOSE_BOX
        wx.Frame.__init__(self, parent, ID, title,
                wx.DefaultPosition, wx.Size(300, default_h),
                wx.DEFAULT_FRAME_STYLE ^ notstyle)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.scenario_chooser = wx.ListBox(self, ID_SCENARIO_CHOOSER)
        sizer.Add(self.scenario_chooser, 1, wx.EXPAND)
        wx.EVT_LISTBOX_DCLICK(self, ID_SCENARIO_CHOOSER, self.on_list_dclick)
        self.SetSizer(sizer)

    def on_scenarios_loaded(self):
        scenario_choices = [app.stringify_scenario(i) for i in app.scenarios]
        self.scenario_chooser.Set(scenario_choices)
        self.Update()

    def on_list_dclick(self, event):
        scen_id = event.GetInt()
        scen = app.scenarios[scen_id]
        app.map_number = scen['map_number']
        squads = (scen['squad1'], scen['squad2'])
        gns_filespec = os.path.join(app.map_path, 'MAP%03i.GNS' % (app.map_number))
        app.viewer.load_traps(app.map_path, app.map_number)
        app.viewer.load_placements(app.map_path, app.map_number, squads)
        if scen['entd'] > 0:
            app.viewer.load_entds(app.map_path, scen['entd'])
        else:
            app.viewer.entds = []
        app.gns_index = app.viewer.load_gns(gns_filespec, (scen['arrangement'], scen['weather'], scen['day_night']))
        if app.texedit_window.IsShown():
            app.texedit_window.canvas.SetCurrent()
            app.texedit.load_map(app.viewer.map, app.texedit_window.canvas.init)
            app.texedit_window.set_button_colors()
            if app.texedit_window.canvas.init:
                app.texedit_window.canvas.on_paint(None)
        app.info_window.on_map_loaded()


class CameraWindow(wx.Frame):
    def __init__(self, parent, ID, title):
        notstyle = wx.MAXIMIZE_BOX | wx.MINIMIZE_BOX | wx.RESIZE_BORDER | wx.CLOSE_BOX
        wx.Frame.__init__(self, parent, ID, title,
                wx.DefaultPosition, wx.Size(300, 300),
                wx.DEFAULT_FRAME_STYLE ^ notstyle)
        btn_size = (40, -1)
        slide_size = (80, -1)
        panel = wx.Panel(self, -1)
        azimuth_staticbox = wx.StaticBox(panel, -1, 'Azimuth')
        altitude_staticbox = wx.StaticBox(panel, -1, 'Altitude')
        magnify_staticbox = wx.StaticBox(panel, -1, 'Magnify')
        fov_staticbox = wx.StaticBox(panel, -1, 'Field of View')
        sizer0 = wx.FlexGridSizer(2, 2)
        sizer_azimuth = wx.StaticBoxSizer(azimuth_staticbox, wx.VERTICAL)
        sizer_azimuth_sub = wx.BoxSizer(wx.HORIZONTAL)
        sizer_altitude = wx.StaticBoxSizer(altitude_staticbox, wx.VERTICAL)
        sizer_altitude_sub = wx.GridSizer(3, 2)
        sizer_magnify = wx.StaticBoxSizer(magnify_staticbox, wx.VERTICAL)
        sizer_fov = wx.StaticBoxSizer(fov_staticbox, wx.VERTICAL)
        self.azimuth_text = wx.TextCtrl(panel, ID_AZIMUTH_TEXT)
        self.azimuth_slide = wx.Slider(panel, ID_AZIMUTH_SLIDE, minValue = 0, maxValue = 360, style = wx.SP_WRAP, size = slide_size)
        self.azimuth_auto = wx.ToggleButton(panel, ID_AZIMUTH_AUTO, 'Auto-Spin')
        self.azimuth_0 = wx.Button(panel, ID_AZIMUTH_0, u'0°')
        self.azimuth_m45 = wx.Button(panel, ID_AZIMUTH_M45, u'-45°', size = btn_size)
        self.azimuth_p45 = wx.Button(panel, ID_AZIMUTH_P45, u'+45°', size = btn_size)
        self.altitude_text = wx.TextCtrl(panel, ID_ALTITUDE_TEXT)
        self.altitude_slide = wx.Slider(panel, ID_ALTITUDE_SLIDE, minValue = 0, maxValue = 90, size = slide_size)
        self.altitude_0 = wx.Button(panel, ID_ALTITUDE_0, u'0°', size = btn_size)
        self.altitude_26 = wx.Button(panel, ID_ALTITUDE_26, u'26°', size = btn_size)
        self.altitude_30 = wx.Button(panel, ID_ALTITUDE_30, u'30°', size = btn_size)
        self.altitude_38 = wx.Button(panel, ID_ALTITUDE_38, u'38°', size = btn_size)
        self.altitude_45 = wx.Button(panel, ID_ALTITUDE_45, u'45°', size = btn_size)
        self.altitude_90 = wx.Button(panel, ID_ALTITUDE_90, u'90°', size = btn_size)
        self.magnify_text = wx.TextCtrl(panel, ID_MAGNIFY_TEXT)
        self.magnify_slide = wx.Slider(panel, ID_MAGNIFY_SLIDE, minValue=1, maxValue=1000, size = slide_size)
        self.magnify_auto = wx.ToggleButton(panel, ID_MAGNIFY_AUTO, 'Fit Width')
        self.magnify_100 = wx.Button(panel, ID_MAGNIFY_100, '100%')
        self.fov_text = wx.TextCtrl(panel, ID_FOV_TEXT)
        self.fov_slide = wx.Slider(panel, ID_FOV_SLIDE, size = slide_size)
        self.fov_isometric = wx.ToggleButton(panel, ID_FOV_ISOMETRIC, 'Isometric')
        self.fov_27 = wx.Button(panel, ID_FOV_27, u'27°')
        self.set_toggle_buttons()
        sizer_azimuth.Add(self.azimuth_text, 0)
        sizer_azimuth.Add(self.azimuth_slide, 0)
        sizer_azimuth.Add(self.azimuth_auto, 0)
        sizer_azimuth.Add(self.azimuth_0, 0)
        sizer_azimuth_sub.Add(self.azimuth_m45, 0)
        sizer_azimuth_sub.Add(self.azimuth_p45, 0)
        sizer_azimuth.Add(sizer_azimuth_sub, 0)
        sizer_altitude.Add(self.altitude_text, 0)
        sizer_altitude.Add(self.altitude_slide, 0)
        sizer_altitude_sub.Add(self.altitude_0, 0)
        sizer_altitude_sub.Add(self.altitude_26, 0)
        sizer_altitude_sub.Add(self.altitude_30, 0)
        sizer_altitude_sub.Add(self.altitude_38, 0)
        sizer_altitude_sub.Add(self.altitude_45, 0)
        sizer_altitude_sub.Add(self.altitude_90, 0)
        sizer_altitude.Add(sizer_altitude_sub, 0)
        sizer_magnify.Add(self.magnify_text, 0)
        sizer_magnify.Add(self.magnify_slide, 0)
        sizer_magnify.Add(self.magnify_auto, 0)
        sizer_magnify.Add(self.magnify_100, 0)
        sizer_fov.Add(self.fov_text, 0)
        sizer_fov.Add(self.fov_slide, 0)
        sizer_fov.Add(self.fov_isometric, 0)
        sizer_fov.Add(self.fov_27, 0)
        sizer0.Add(sizer_azimuth, 0)
        sizer0.Add(sizer_altitude, 0)
        sizer0.Add(sizer_magnify, 0)
        sizer0.Add(sizer_fov, 0)
        sizer0.Fit(panel)
        sizer0.Fit(self)
        self.azimuth_slide.Bind(wx.EVT_SCROLL, self.on_controls)
        self.altitude_slide.Bind(wx.EVT_SCROLL, self.on_controls)
        self.magnify_slide.Bind(wx.EVT_SCROLL, self.on_controls)
        self.fov_slide.Bind(wx.EVT_SCROLL, self.on_controls)
        wx.EVT_TOGGLEBUTTON(self, ID_AZIMUTH_AUTO, self.on_controls)
        wx.EVT_TOGGLEBUTTON(self, ID_MAGNIFY_AUTO, self.on_controls)
        wx.EVT_TOGGLEBUTTON(self, ID_FOV_ISOMETRIC, self.on_controls)
        wx.EVT_BUTTON(self, ID_AZIMUTH_0, self.on_controls)
        wx.EVT_BUTTON(self, ID_AZIMUTH_P45, self.on_controls)
        wx.EVT_BUTTON(self, ID_AZIMUTH_M45, self.on_controls)
        wx.EVT_BUTTON(self, ID_ALTITUDE_0, self.on_controls)
        wx.EVT_BUTTON(self, ID_ALTITUDE_26, self.on_controls)
        wx.EVT_BUTTON(self, ID_ALTITUDE_30, self.on_controls)
        wx.EVT_BUTTON(self, ID_ALTITUDE_38, self.on_controls)
        wx.EVT_BUTTON(self, ID_ALTITUDE_45, self.on_controls)
        wx.EVT_BUTTON(self, ID_ALTITUDE_90, self.on_controls)
        wx.EVT_BUTTON(self, ID_MAGNIFY_100, self.on_controls)
        wx.EVT_BUTTON(self, ID_FOV_27, self.on_controls)
        self.SetSizer(sizer0)

    def on_controls(self, event):
        eid = event.GetId()
        if eid == ID_AZIMUTH_SLIDE: app.viewer.set_azimuth(event.GetPosition())
        elif eid == ID_AZIMUTH_AUTO: app.viewer.set_azimuth_auto(event)
        elif eid == ID_ALTITUDE_SLIDE: app.viewer.set_altitude(event.GetPosition())
        elif eid == ID_MAGNIFY_SLIDE: app.viewer.set_magnify(event.GetPosition())
        elif eid == ID_FOV_SLIDE: app.viewer.set_fov(event.GetPosition())
        elif eid == ID_FOV_ISOMETRIC: app.viewer.set_fov_isometric(event)
        elif eid == ID_AZIMUTH_0: app.viewer.set_azimuth(0)
        elif eid == ID_AZIMUTH_M45: app.viewer.snap_azimuth(-45)
        elif eid == ID_AZIMUTH_P45: app.viewer.snap_azimuth(45)
        elif eid == ID_ALTITUDE_0: app.viewer.set_altitude(0)
        elif eid == ID_ALTITUDE_26: app.viewer.set_altitude(26)
        elif eid == ID_ALTITUDE_30: app.viewer.set_altitude(30)
        elif eid == ID_ALTITUDE_38: app.viewer.set_altitude(38)
        elif eid == ID_ALTITUDE_45: app.viewer.set_altitude(45)
        elif eid == ID_ALTITUDE_90: app.viewer.set_altitude(90)
        elif eid == ID_MAGNIFY_AUTO: app.viewer.set_magnify(None)
        elif eid == ID_MAGNIFY_100: app.viewer.set_magnify(100)
        elif eid == ID_FOV_27: app.viewer.set_fov(27)
        app.camera_window.set_toggle_buttons()

    def set_toggle_buttons(self):
        self.azimuth_auto.SetValue(not app.viewer.angleinc == 0)
        self.magnify_auto.SetValue(app.viewer.magnify == None)
        self.fov_isometric.SetValue(app.viewer.use_ortho)

    def on_values_changed(self):
        self.azimuth_text.SetValue(str(app.viewer.angle))
        self.azimuth_slide.SetValue(int(app.viewer.angle))
        self.altitude_text.SetValue(str(app.viewer.altitude))
        self.altitude_slide.SetValue(int(app.viewer.altitude))
        if app.viewer.magnify == None:
            magnify = int(app.viewer.get_magnify())
            self.magnify_text.ChangeValue(str(magnify))
            self.magnify_slide.SetValue(magnify)
        else:
            self.magnify_text.ChangeValue(str(int(app.viewer.magnify)))
            self.magnify_slide.SetValue(int(app.viewer.magnify))
        self.fov_text.SetValue(str(app.viewer.fov))
        self.fov_slide.SetValue(int(app.viewer.fov))
        self.Update()


class TexEditCanvas(wx.glcanvas.GLCanvas):
    def __init__(self, parent, name, size=(512, 512)):
        attribs = [wx.glcanvas.WX_GL_RGBA, wx.glcanvas.WX_GL_DOUBLEBUFFER, wx.glcanvas.WX_GL_DEPTH_SIZE, 24]
        wx.glcanvas.GLCanvas.__init__(self, parent, -1, name = name, size = size, attribList=attribs)
        self.init = False
        wx.EVT_PAINT(self, self.on_paint)
        wx.EVT_SIZE(self, self.on_resize)
        wx.EVT_ERASE_BACKGROUND(self, self.on_erase_bg)

    def on_resize(self, event):
        size = self.GetClientSize()
        (self.W, self.H) = (size.width, size.height)
        if self.GetContext():
            self.SetCurrent()
            app.texedit.gl_set_viewport(size.width, size.height)
        event.Skip()

    def on_paint(self, event):
        if event:
            dc = wx.PaintDC(self)
        self.SetCurrent()
        if not self.init:
            self.init = True
            app.texedit.gl_initialize()
            size = self.GetClientSize()
            (self.W, self.H) = (size.width, size.height)
            app.texedit.gl_set_viewport(self.W, self.H)
        self.draw_gl_stuff()

    def draw_gl_stuff(self):
        app.texedit.gl_draw()
        self.SwapBuffers()

    def on_erase_bg(self, event):
        # Prevent flickering in Windows
        pass


class TexEditWindow(wx.Frame):
    def __init__(self, parent, ID, title):
        wx.Frame.__init__(self, parent, ID, title,
                wx.DefaultPosition)
        self.SetClientSize((960, 560))
        self.create_menu_bar()

        panel = wx.Panel(self, -1)
        sizer_panes = wx.BoxSizer(wx.HORIZONTAL)
        sizer_palettes = wx.BoxSizer(wx.VERTICAL)
        sizer_color_sliders = wx.BoxSizer(wx.VERTICAL)
        self.show_used = wx.CheckBox(panel, ID_SHOW_USED_AREAS, 'Show areas used by polygons')
        self.show_used.SetValue(True)
        sizer_palettes.Add(self.show_used, 0)
        wx.EVT_CHECKBOX(self, ID_SHOW_USED_AREAS, self.toggle_show_used)
        self.palette_radios = []
        self.color_buttons = []
        self.active_button = None
        self.color_sliders = []
        self.color_inputs = []
        for i in range(16):
            sizer_colors = wx.BoxSizer(wx.HORIZONTAL)
            radio = wx.RadioButton(panel, ID_PALETTE_RADIO_0 + i, 'Palette %02u' % i)
            wx.EVT_RADIOBUTTON(self, ID_PALETTE_RADIO_0 + i, self.on_radio_click)
            sizer_colors.Add(radio, 0)
            self.palette_radios.append(radio)
            for j in range(16):
                color_button = wx.Button(panel,
                        ID_COLOR_BUTTON_0 + i * 16 + j, '*',
                        size = wx.Size(20, 20)
                )
                color_button.Bind(wx.EVT_BUTTON, self.on_color_button)
                self.color_buttons.append(color_button)
                sizer_colors.Add(color_button, 0)
            sizer_palettes.Add(sizer_colors, 0)
        for i, color in enumerate(['R', 'G', 'B', 'A']):
            sizer_slide_input = wx.BoxSizer(wx.HORIZONTAL)
            max_value = 31
            if color == 'A':
                max_value = 1
            color_label = wx.StaticText(panel, ID_COLOR_LABEL_0 + i, color, size=wx.Size(20, -1))
            color_slider = wx.Slider(panel, ID_COLOR_SLIDER_0 + i, 0, 0, max_value, size=wx.Size(310, -1))
            color_slider.Bind(wx.EVT_SCROLL, self.on_color_slide)
            color_input = wx.TextCtrl(panel, ID_COLOR_INPUT_0 + i)
            sizer_slide_input.Add(color_label, 0)
            sizer_slide_input.Add(color_slider, 0)
            sizer_slide_input.Add(color_input, 0)
            sizer_color_sliders.Add(sizer_slide_input)
            self.color_sliders.append(color_slider)
            self.color_inputs.append(color_input)
        sizer_palettes.Add(sizer_color_sliders, 0)
        self.canvas = TexEditCanvas(panel, 'Texture Editor')
        sizer_panes.Add(self.canvas, 0)
        sizer_panes.Add(sizer_palettes, 0)
        sizer_panes.Fit(panel)
        sizer_panes.Fit(self)
        self.SetSizer(sizer_panes)

    def on_color_button(self, event):
        button = event.GetEventObject()
        print 'setting active button', button
        self.active_button = button

    def on_color_slide(self, event):
        slider = event.GetId()
        position = event.GetPosition()
        color = slider - ID_COLOR_SLIDER_0
        color_input = self.color_inputs[color]
        color_input.SetValue(str(position))
        color_input.Update()
        if not self.active_button:
            return
        factor = 255.0 / 31.0
        r = round(self.color_sliders[0].GetValue() * factor)
        g = round(self.color_sliders[1].GetValue() * factor)
        b = round(self.color_sliders[2].GetValue() * factor)
        a = self.color_sliders[3].GetValue() * 255
        self.set_button_color(self.active_button, r, g, b, a)

    def set_button_color(self, button, r, g, b, a):
        back = wx.Colour(r, g, b)
        fore = back
        if a == 0:
            white = wx.Colour(255, 255, 255)
            fore = white
        button.SetOwnBackgroundColour(back)
        button.SetOwnForegroundColour(fore)

    def set_button_colors(self):
        factor = 255.0 / 31.0
        palettes = list(app.texedit.map.palettes_color)
        alphas = list(app.texedit.map.palettes_color_alpha)
        for i, palette in enumerate(palettes):
            for j in range(16):
                r = round(palette[j * 3 + 0] * factor)
                g = round(palette[j * 3 + 1] * factor)
                b = round(palette[j * 3 + 2] * factor)
                a = alphas[i][j*3]
                self.set_button_color(self.color_buttons[i * 16 + j], r, g, b, a)

    def toggle_show_used(self, event):
        check = event.GetEventObject()
        state = check.GetValue()
        app.texedit_window.canvas.SetCurrent()
        app.texedit.set_show_used(state)
        app.texedit_window.canvas.on_paint(None)

    def on_radio_click(self, event):
        radio = event.GetEventObject()
        text = radio.GetLabelText()
        index = int(text[8:])
        app.texedit_window.canvas.SetCurrent()
        app.texedit.gl_set_palette(index)
        app.texedit_window.canvas.on_paint(None)

    def create_menu_bar(self):
        menuBar = wx.MenuBar()
        menu = wx.Menu()
        menu.Append(ID_EXPORT, "&Export", "Export this texture as a PNG file")
        menu.Append(ID_IMPORT, "&Import", "Import a PNG file to replace this texture")
        menuBar.Append(menu, "&File")
        self.SetMenuBar(menuBar)
        wx.EVT_MENU(self, ID_EXPORT, self.export_file)
        wx.EVT_MENU(self, ID_IMPORT, self.import_file)

    def export_file(self, event):
        texfile_name = os.path.basename(app.texedit.map.texture)
        palette_number = app.texedit.palette
        default_name = '%s-pal%02u.png' % (texfile_name, palette_number)
        dlg = wx.FileDialog(self, "Export texture to file", defaultDir = '', defaultFile = default_name, wildcard = '*.png')
        if dlg.ShowModal() == wx.ID_OK:
            file_spec = dlg.GetPath()
            if file_spec[-4:].lower() != '.png':
                file_spec += '.png'
            app.texedit.export_tex(file_spec)
        dlg.Destroy()

    def import_file(self, event):
        dlg = wx.FileDialog(self, "Import texture from file", wildcard = '*.png')
        if dlg.ShowModal() == wx.ID_OK:
            file_spec = dlg.GetPath()
            app.texedit.import_tex(file_spec)
        dlg.Destroy()
        gns_filespec = os.path.join(app.map_path, 'MAP%03i.GNS' % (app.map_number))
        app.viewer.load_gns(gns_filespec, app.gns_index)
        app.texedit_window.canvas.SetCurrent()
        app.texedit.load_map(app.viewer.map, app.texedit_window.canvas.init)
        app.texedit_window.set_button_colors()
        if app.texedit_window.canvas.init:
            app.texedit_window.canvas.on_paint(None)


class MyApp(wx.App):
    def OnInit(self):
        self.scenarios = None
        return True

    def _init(self):
        # File/path settings
        self.map_path = None
        self.map_number = 0
        self.gns_index = 0
        self.viewer = fft.map.viewer.Viewer()
        self.viewer_window = ViewerWindow(None, -1, 'map2gl')
        self.viewer_window.Show(True)
        self.scenario_window = ScenarioWindow(self.viewer_window, -1, 'Scenarios')
        self.info_window = InfoWindow(self.viewer_window, -1, 'Information')
        self.camera_window = CameraWindow(self.viewer_window, -1, 'Camera')
        self.texedit = fft.map.texedit.TexEdit()
        self.texedit_window = TexEditWindow(self.viewer_window, -1, 'Texture Editor')
        size = self.viewer_window.GetClientSize()
        self.canvas = ViewerCanvas(self.viewer_window, 'map2gl', size)
        wx.EVT_CHAR(self, self.on_char)
        self.SetTopWindow(self.viewer_window)
        app.canvas.on_paint(None)
        if len(sys.argv) > 1:
            self.viewer_window.open_file(sys.argv[1])

    def get_scenarios(self):
        F = fft.AttackOut(os.path.join(app.map_path, '..', 'EVENT', 'ATTACK.OUT'))
        self.scenarios = F.read_scenarios()
        self.scenario_window.on_scenarios_loaded()

    def stringify_scenario(self, scen):
        return fft.map.names[scen['map_number']]

    def on_char(self, event):
        key = chr(event.GetKeyCode())

        # i raises camera
        # k lowers camera
        if key == 'k':
            app.viewer.altitude = round(app.viewer.altitude / 2) * 2 - 2
        if key == 'i':
            app.viewer.altitude = round(app.viewer.altitude / 2) * 2 + 2

        # j finds next even 2deg viewpoint to the left
        # l finds next even 2deg viewpoint to the right
        if key == 'j': app.viewer.snap_azimuth(-2)
        if key == 'l': app.viewer.snap_azimuth(2)

        # I raises camera
        # K lowers camera
        # 26 and 38 seem to be the in-game low and high angles
        if key == 'K':
            if app.viewer.altitude <= 30: app.viewer.altitude = 0
            elif app.viewer.altitude <= 45: app.viewer.altitude = 30
            elif app.viewer.altitude <= 60: app.viewer.altitude = 45
            elif app.viewer.altitude <= 90: app.viewer.altitude = 60
            else: app.viewer.altitude = 90
        if key == 'I':
            if app.viewer.altitude >= 60: app.viewer.altitude = 90
            elif app.viewer.altitude >= 45: app.viewer.altitude = 60
            elif app.viewer.altitude >= 30: app.viewer.altitude = 45
            elif app.viewer.altitude >= 0: app.viewer.altitude = 30
            else: app.viewer.altitude = 0

        # J finds next even 45deg viewpoint to the left
        # L finds next even 45deg viewpoint to the right
        if key == 'J': app.viewer.snap_azimuth(-45)
        if key == 'L': app.viewer.snap_azimuth(45)

        # , accelerates leftward rotation
        # . accelerates rightward rotation
        if key == ',': app.viewer.angleinc -= .01
        if key == '.': app.viewer.angleinc += .01

        # = moves camera closer
        # - moves camera farther away
        if (key == '-' or key == '=') and app.viewer.magnify == None:
            app.viewer.magnify = app.viewer.get_magnify()
        if key == '-': app.viewer.magnify /= 1.1
        if key == '=': app.viewer.magnify *= 1.1
        if (key == '-' or key == '='):
            app.viewer.set_orthobase()

        # + increases focal length (perspective projection only)
        # _ decreases focal length (perspective projection only)
        if not app.viewer.use_ortho:
            if key == '+': app.viewer.fov -= 1
            if key == '_': app.viewer.fov += 1

        # s toggles skirt on/off
        if key == 's':
            app.viewer.use_skirt = not app.viewer.use_skirt
            app.viewer.create_mesh_dlists()
            app.viewer.create_mesh_anim_dlists()

        # c toggles cutaway on/off
        if key == 'c':
            app.viewer.use_cutaway = not app.viewer.use_cutaway
            app.viewer.create_mesh_dlists()
            app.viewer.create_mesh_anim_dlists()

        # g toggles lights on/off
        if key == 'g':
            app.viewer.gl_set_use_lights(not app.viewer.use_lights)

        # B toggles background chroma key
        if key == 'B': app.viewer.chroma_key = not app.viewer.chroma_key

        # o toggles orthographic (isometric) projection
        if key == 'o': app.viewer.use_ortho = not app.viewer.use_ortho

        # x toggles display of axes
        if key == 'x': app.viewer.show_axes = not app.viewer.show_axes

        # t toggles display of terrain
        if key == 't': app.viewer.use_terrain = not app.viewer.use_terrain

        # t toggles display of textures
        if key == 'T': app.viewer.gl_set_use_texture(not app.viewer.use_texture)

        # [ selects previous map
        # ] selects next map
        if key == '[': app.map_number -= 1
        if key == ']': app.map_number += 1

        # 0 resets vertical and horizontal viewing angle to 0 and stops rotation
        if key == '0':
            app.viewer.angle = 0
            app.viewer.angleinc = 0
            app.viewer.altitude = 0

        # a toggles animation of extra meshes
        if key == 'a':
            app.viewer.do_animation = not app.viewer.do_animation
            if not app.viewer.do_animation: app.viewer.anim_timer = 0

        # z sets zoom level to in-game zoom level
        if key == 'z':
            app.viewer.magnify = 100
            app.viewer.set_orthobase()

        # re-init the camera perspective if necessary
        if '-=_+oz'.find(key) > -1: app.viewer.gl_set_perspective()

        if key == '{':
            app.gns_index -= 1
            app.gns_index %= len(app.viewer.map.all_indexes)

        if key == '}':
            app.gns_index += 1
            app.gns_index %= len(app.viewer.map.all_indexes)

        # Load next or previous map if [ or ] was pressed
        if key == '[' or key == ']' or key == '{' or key == '}':
            if key == '[' or key == ']':
                app.gns_index = 0
            gns_filespec = os.path.join(app.map_path, 'MAP%03i.GNS' % (app.map_number))
            app.viewer.load_traps(app.map_path, app.map_number)
            app.viewer.load_placements(app.map_path, app.map_number)
            app.viewer.entds = []
            app.viewer.load_gns(gns_filespec, app.gns_index)
            if app.texedit_window.IsShown():
                app.texedit_window.canvas.SetCurrent()
                app.texedit.load_map(app.viewer.map, app.texedit_window.canvas.init)
                app.texedit_window.set_button_colors()
                if app.texedit_window.canvas.init:
                    app.texedit_window.canvas.on_paint(None)
            self.info_window.on_map_loaded()

        # If escape is pressed, kill everything.
        if ord(key) == wx.WXK_ESCAPE:
            sys.exit()

        if 'ijklIJKL0'.find(key) > -1:
            app.viewer.mod_angles()

        app.camera_window.set_toggle_buttons()

app = MyApp(0)
app._init()
if not app.viewer.map_is_loaded:
    app.viewer_window.open_file_dialog(None)
app.MainLoop()
